package denim;

import beast.base.core.BEASTObject;
import beast.base.core.Description;
import beast.base.core.Input;
import beast.base.core.Loggable;
import beast.base.evolution.tree.Node;
import beast.base.evolution.tree.Tree;
import beast.base.evolution.tree.TreeInterface;
import denim.util.STreeBranchIncident;

import java.io.PrintStream;
import java.math.RoundingMode;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

/*
 * Created by GRJ on 05/06/2017.
 * Cut-down version of TreeWithMetaDataLogger, with my own bits added
 */

@Description("Logger which store information about migrations as annotations in the MCMC samples of species tree.")

public class SpeciesTreeWithMigrationsLogger extends BEASTObject implements Loggable {

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    final public Input<CoalescenceMigrationDistribution> coalMigDistInput =
            new Input<>("coalMigDist",
                    "The coalescent-migration distribution.", Input.Validate.REQUIRED);

    @Override
    public void initAndValidate() {
    }

    @Override
    public void init(PrintStream out) {
        Tree sTree = coalMigDistInput.get().getStree();
        sTree.init(out);
        CoalescenceMigrationDistribution cmd = coalMigDistInput.get();
        out.print("\n[DENIM_mapping_from_indices_to_alignment_IDs ");
        for (int j = 0; j < cmd.gTreeWEsInput.get().size(); j++) {
            String gtID = cmd.getAlignmentID(j);
            out.print(j + "=" + gtID);
            if (j+1 < cmd.gTreeWEsInput.get().size()) {
                out.print(" ");
            }
        }
        out.print("]");
    }

    @Override
    public void log(long nSample, PrintStream out) {
        CoalescenceMigrationDistribution cmd = coalMigDistInput.get();
        cmd.updateMigrationsHeights();
        Tree sTree = cmd.getStree();
        // write out the log tree with meta data
        out.print("tree STATE_" + nSample + " = ");
        sTree.getRoot().sort();
        out.print(toNewick(sTree.getRoot(), cmd));
        //out.print(tree.getRoot().toShortNewick(false));
        out.print(";");
    }




    private String toNewick(Node node, CoalescenceMigrationDistribution cmd) {
        StringBuffer buf = new StringBuffer();
        Node lft = node.getChildCount() > 0 ? node.getChild(0) : null;
        Node rgt = node.getChildCount() > 1 ? node.getChild(1) : null;
        if (lft != null) {
            buf.append("(");
            buf.append(toNewick(lft, cmd));
            if (rgt != null) {
                buf.append(',');
                buf.append(toNewick(rgt, cmd));
            }
            buf.append(")");
        } else {
            buf.append(node.getNr() + 1);
        }
        ArrayList<ArrayList<STreeBranchIncident>> migs = cmd.getMigrationsInBranch(node);
        int nMigs = 0;
        for (int j = 0 ; j < migs.size(); j++) {
            nMigs += migs.get(j).size();
        }
        if (nMigs > 0) {
            buf.append("[&");
            buf.append("migs={");
            int n = 0;
            for (int j = 0 ; j < migs.size(); j++) {
                for (STreeBranchIncident sbi : migs.get(j)) {
                    int ch = sbi.getChange();
                    double hgt = sbi.getHeight();
                    //String locus = "X" + cmd.getAlignmentID(j);
                    buf.append(j + "," + ch + "," + hgt);
                    n += 1;
                    if (n < nMigs) {
                        buf.append(",");
                    }
                }
            }
            buf.append("}");
            buf.append(']');
        }
        buf.append(":");

        double nodeLength;
        if (node.isRoot()) {
            nodeLength = cmd.getMaxTreeHeight() - node.getHeight();
        } else {
            nodeLength = node.getLength();
        }
        buf.append(nodeLength);

        return buf.toString();
    }


    @Override
    public void close(PrintStream out) {
        Tree sTree = coalMigDistInput.get().getStree();
        sTree.close(out);
    }


}
