package denim.util;

import beast.base.core.Description;

import static denim.util.STreeBranchIncident.Type.*;

/*
 * Created by Work on 27/10/2016.
 */

/*
STreeBranchIncident's are used to work along a branch accumulating total intensity of coalescence
or of migration from this branch to another (specific) branch
    */
@Description("A small value class which stores an incident along a species tree branch.")

public class STreeBranchIncident {
    private double height;
    private int change; // 0 (for ends of stree branches), +1 (mig arrives) or -1 (mig leaves or coal)
    private Type type;
    private int destNr;
    private int epoch;


    public enum Type {
        COALESCENCE,
        MIGRATION_IN,
        MIGRATION_OUT,
        BRANCH_START,
        EPOCH_BOUNDARY,
        BRANCH_END
    }



    private STreeBranchIncident(Type type) {
        this.height = -1.0;
        this.change = 0;
        this.destNr = -1;
        this.epoch = -1;
        this.type = type;
    }


    static STreeBranchIncident makeCoalescenceIncident(double height) {
        STreeBranchIncident sbi = new STreeBranchIncident(COALESCENCE);
        sbi.height = height;
        sbi.change = -1;
        return sbi;
    }


    static STreeBranchIncident makeInMigrationIncident(double height) {
        STreeBranchIncident sbi = new STreeBranchIncident(MIGRATION_IN);
        sbi.height = height;
        sbi.change = 1;
        return sbi;
    }


    static STreeBranchIncident makeOutMigrationIncident(double height, int destNr) {
        STreeBranchIncident sbi = new STreeBranchIncident(MIGRATION_OUT);
        sbi.height = height;
        sbi.change = -1;
        sbi.destNr = destNr;
        return sbi;
    }


    static STreeBranchIncident makeBranchStart(double height) {
        return makeBoundaryIncident(height, BRANCH_START);
    }

    static STreeBranchIncident makeBranchEnd(double height) {
        return makeBoundaryIncident(height, BRANCH_END);
    }


    static STreeBranchIncident makeEpochIncident(double height, int epoch) {
        STreeBranchIncident sbi = makeBoundaryIncident(height, EPOCH_BOUNDARY);
        sbi.epoch = epoch;
        return sbi;
    }


    public double getHeight() {
        return height;
    }

    public int getChange () {
        return change;
    }

    public Type getType () {
        return type;
    }

    public int getDestNr() {
        return destNr;
    }

    public int getEpoch() {
        return epoch;
    }


    public static int compareTypes(STreeBranchIncident x, STreeBranchIncident y) {
        return x.type.compareTo(y.type);
    }



    /***********************************************************************************/

    private static STreeBranchIncident makeBoundaryIncident(double height, Type type) {
        STreeBranchIncident sbi = new STreeBranchIncident(type);
        sbi.height = height;
        return sbi;
    }

}
